/***************************************************************************
 *   Copyright (C) 2003 by Sbastien Laot                                 *
 *   sebastien.laout@tuxfamily.org                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qbuffer.h>
#include <kcolordrag.h>
#include <kapplication.h>
#include <kiconloader.h>
#include <qimage.h>
#include <kurldrag.h>
#include <qtimer.h>

#include "itemdrag.h"
#include "itemfactory.h"
#include "basket.h"

/** ItemDrag */

const char * ItemDrag::item_mime_string  = "application/x-basket-item";

ItemDrag::ItemDrag(Item *item, QWidget *parent, const char *name)
 : QDragObject(parent, name)
{
	/* Init item */
	m_type        = item->type();
	m_isChecked   = item->isChecked();
	m_annotations = item->annotations();
	m_fileName    = item->fileName();

	/* Init item content */
	switch (m_type) {
		case Item::Text:
			m_text         = item->text();
			m_textFontType = item->textFontType();
			m_color        = item->textColor();
			break;
		case Item::Html:
			m_text         = item->html();
			m_showSource   = item->showSource();
			break;
		case Item::Image:
			m_pixmap       = QPixmap(*(item->pixmap()));
			break;
		case Item::Link:
			m_url          = item->url();
			m_title        = item->title();
			m_icon         = item->icon();
			m_autoTitle    = item->autoTitle();
			m_autoIcon     = item->autoIcon();
			break;
		case Item::Color:
			m_color        = item->color();
			break;
		case Item::Animation:
		case Item::Sound:
		case Item::File:
		case Item::Launcher:
			break;
	}
}

ItemDrag::~ItemDrag()
{
}

const char* ItemDrag::format(int i) const
{
	if (i == 0)
		return item_mime_string;
	else
		return 0;
}

QByteArray ItemDrag::encodedData(const char *mimeType) const
{
	if ( ::qstrcmp(mimeType, item_mime_string) != 0 )
		return QByteArray(); // Return if it isn't the BasketItem MIME string

	QBuffer buffer;

	if (buffer.open(IO_WriteOnly)) {
		QDataStream stream(&buffer);
		stream << (Q_INT32)m_type
		       << (Q_INT32)m_isChecked // Basket::trueOrFalse() : "true" / "false" ? Because I don't know how to store bools
		       << m_annotations
		       << m_fileName;
		switch (m_type) {
			case Item::Text:
				stream << m_text
				       << m_textFontType
				       << m_color;
				break;
			case Item::Html:
				stream << m_text
				       << (Q_INT32)m_showSource;
				break;
			case Item::Image:
				stream << m_pixmap;
				break;
			case Item::Link:
				stream << m_url
				       << m_title
				       << m_icon
				       << (Q_INT32)m_autoTitle
				       << (Q_INT32)m_autoIcon;
				break;
			case Item::Color:
				stream << m_color;
				break;
		case Item::Animation:
		case Item::Sound:
		case Item::File:
		case Item::Launcher:
			break; // We just need fileName is those cases
		}
		buffer.close();
	}

	return buffer.buffer();
}

bool ItemDrag::canDecode(QMimeSource *source)
{
	return (source != 0) && (source->provides(item_mime_string));
}

Item* ItemDrag::decode(QMimeSource *source, Basket *parent)
{
	/* Variables to handle item type and global properties */
	Item::Type type;
	Q_INT32    tmp_isChecked;
	bool       isChecked;
	QString    annotations;
	QString    fileName;

	/* Special variables to handle contents by type */
	QString text;
	int     textFontType;
	Q_INT32 showSource;
	QPixmap pixmap;
	QColor  color;
	KURL    url;
	QString title;
	QString icon;
	Q_INT32 autoTitle;
	Q_INT32 autoIcon;

	bool shouldCopy = false;

	/* Decode the item */
	QBuffer buffer(source->encodedData(item_mime_string));
	Item *item = 0;

	if (buffer.open(IO_ReadOnly)) {
		QDataStream stream(&buffer);
		stream >> (Q_INT32&)type // FIXME: Alwayse Q_INT32 ?
		       >> tmp_isChecked
		       >> annotations
		       >> fileName;
		isChecked = tmp_isChecked != 0;
		if ( Item::isAMirror(fileName) && parent->itemForFullPath(fileName) ) {
			shouldCopy = true;
			// Notice it to the user as soon as the drop is processed :
			parent->m_mirrorOnlyOnceFileName = fileName;
			QTimer::singleShot( 0, parent, SLOT(showMirrorOnlyOnceInfo()) ); // TODO: Show also in statusBar ?
		}
		switch (type) {
			case Item::Text:
				stream >> text
				       >> textFontType
				       >> color;
				if (Item::isAMirror(fileName)) {
					item = new Item( fileName, textFontType, color,
					                 annotations, isChecked, parent );
					item->loadContent();
					if (shouldCopy)       // Hey : I'm lazy and 0.5.0alpha(s) are a lot of works !
						item->unmirror(); // But it works ;-)
				} else {
					item = new Item( ItemFactory::createFileForNewItem(parent, "txt", fileName), textFontType, color,
					                 annotations, isChecked, parent );
					item->setText(text);
				}
				break;
			case Item::Html:
				stream >> text
				       >> showSource;
				if (Item::isAMirror(fileName)) {
					item = new Item( fileName, showSource != 0,
					                 annotations, isChecked, parent );
					item->loadContent();
					if (shouldCopy)
						item->unmirror();
				} else {
					item = new Item( ItemFactory::createFileForNewItem(parent, "html", fileName), showSource != 0,
					                 annotations, isChecked, parent );
					item->setHtml(text);
				}
				break;
			case Item::Image:
				if (Item::isAMirror(fileName)) {
					item = new Item( fileName, Item::Image,
					                 annotations, isChecked, parent );
					item->loadContent();
					if (shouldCopy)
						item->unmirror();
				} else {
					stream >> pixmap;
					item = new Item( ItemFactory::createFileForNewItem(parent, "png", fileName),
					                 annotations, isChecked, parent );
					item->setPixmap(pixmap);
				}
				break;
			case Item::Link:
				stream >> url
				       >> title
				       >> icon
				       >> autoTitle
				       >> autoIcon;
				item = new Item( url, title, icon,
				                 autoTitle != 0, autoIcon != 0,
				                 annotations, isChecked, parent );
				break;
			case Item::Color:
				stream >> color;
				item = new Item( color,
				                 annotations, isChecked, parent );
				break;
		}
		buffer.close();
	}

	return item;
}

/** ItemMultipleDrag */

ItemMultipleDrag::ItemMultipleDrag(Item *item, QWidget *parent, const char *name)
 : KMultipleDrag(parent, name)
{
	KColorDrag *cDrag = 0;

	/* Main format for item content */
	addDragObject( new ItemDrag(item, 0) );

	/* Alternative formats for item content */
	switch (item->type()) {
		case Item::Text:  addDragObject( new QTextDrag(  item->text(),                     0     ) ); break;
		case Item::Html:  addDragObject( new QTextDrag(  item->html(),                     0     ) ); break;
		case Item::Image: addDragObject( new QImageDrag( item->pixmap()->convertToImage(), 0     ) ); break;
		case Item::Link:  addDragObject( new KURLDrag(   KURL::List(item->url()),          0, "" ) ); break;
		case Item::Color: addDragObject( cDrag = new KColorDrag( item->color(),            0     ) );
		                  addDragObject( new QTextDrag(  item->color().name(),             0     ) ); break;
		case Item::Animation:
		case Item::Sound:
		case Item::File:
		case Item::Launcher:
			addDragObject( new KURLDrag(KURL::List(KURL(item->fullPath())), 0, "") );
			break;
	}

	/* Alternative format (the path) for items stored in files */
	if (item->useFile())
		addDragObject( new KURLDrag(KURL(item->fullPath()), 0, "") );

	/* Init icon that is shown during drag */
	QPixmap      icon;
	QImage       image; // For temporary scale purpose
	const int    iconSize = 22;
	const QPoint iconHotPoint = QPoint(-5, -7); //iconSize / 2;

	if (1/*TODO*/)
		switch (item->type()) {
			case Item::Text:
				icon = kapp->iconLoader()->loadIcon("text", KIcon::NoGroup, iconSize);
				setPixmap(icon, iconHotPoint);
				break;
			case Item::Html:
				icon = kapp->iconLoader()->loadIcon("html", KIcon::NoGroup, iconSize);
				setPixmap(icon, iconHotPoint);
				break;
			case Item::Image:
				image = item->pixmap()->convertToImage();
				image = image.smoothScale(40, 30, QImage::ScaleMin);
				icon = QPixmap(image);
				setPixmap(icon, iconHotPoint);
				break;
			case Item::Link:
				icon = kapp->iconLoader()->loadIcon(item->icon(), KIcon::NoGroup, iconSize);
				setPixmap(icon, iconHotPoint);
				break;
			case Item::Color:
				setPixmap(cDrag->pixmap(), cDrag->pixmapHotSpot());
				break;
			case Item::Animation:
			case Item::Sound:
			case Item::File:
			case Item::Launcher:
				icon = kapp->iconLoader()->loadIcon(ItemFactory::iconForURL(item->fullPath()), KIcon::NoGroup, iconSize);
				setPixmap(icon, iconHotPoint);
				break;
		}
}

ItemMultipleDrag::~ItemMultipleDrag()
{
}

#include "itemdrag.moc"
